/*
 *  CUPS add-on module for Canon Bubble Jet Printer.
 *  Copyright CANON INC. 2001-2004
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <cups/cups.h>
#include <cups/ppd.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>

#include "paramlist.h"
#include "bjexec.h"
#include "canonopt.h"

//#define DEBUG_SLEEP
//#define DEBUG_PS


#define	SHELL_PATH		"/bin/"
#define	SHELL_NAME		"sh"
#define	BUF_SIZE		(1024 * 256)

#define PS_PATH			"/tmp/bj_data%d.ps"
#define BEGIN_FEATURE	"\%\%BeginFeature"
#define PAGE_DEV_BEGIN	"<<"
#define PAGE_DEV_END	">>setpagedevice"

#define IS_BLANK(c)		(c == ' '  || c == '\t')
#define IS_RETURN(c)	(c == '\r' || c == '\n')


static
ParamList *get_ps_params(int *p_ifd, char **p_ps_path)
{
	FILE *ps_fp;
	int ifd = *p_ifd;
	int cc;
	ParamList *p_list = NULL;
	unsigned char *p_read_buf;
	unsigned char *p_data_buf;
	int read_byte;
	static char ps_path[256];
	char key[256];
	char value[256];

	sprintf(ps_path, PS_PATH, getpid());
	ps_fp = fopen(ps_path, "w+");
	*p_ps_path = ps_path;

	if( ps_fp == NULL )
		return NULL;

	p_read_buf = (char*)malloc(BUF_SIZE);
	p_data_buf = (char*)malloc(BUF_SIZE);

	if( p_read_buf != NULL && p_data_buf != NULL )
	{
		while( (read_byte = read(ifd, p_read_buf, BUF_SIZE)) > 0 )
		{
			int read_pos = 0;
			int data_byte = 0;

			while( read_pos < read_byte )
			{
				p_data_buf[data_byte++] = cc = p_read_buf[read_pos++];

				if( data_byte == BUF_SIZE )
				{
					if( fwrite(p_data_buf, 1, data_byte, ps_fp) != data_byte )
						goto write_error;
					data_byte = 0;
				}
				else if( IS_RETURN(cc) )
				{
					int begin_len = strlen(BEGIN_FEATURE);
					int dev_len = strlen(PAGE_DEV_BEGIN);
					int end_len = strlen(PAGE_DEV_END);

					// Interpret %%BeginFeature comment.
					if( strncmp(p_data_buf, BEGIN_FEATURE, begin_len) == 0 )
					{
						int pos = begin_len;
						int key_pos = 0;
						int val_pos = 0;

						while( IS_BLANK(p_data_buf[pos])
							|| p_data_buf[pos] == ':'
							|| p_data_buf[pos] == '*' )	pos++;

						while( !IS_BLANK(p_data_buf[pos])
							&& !IS_RETURN(p_data_buf[pos])
							&& key_pos < 255 )
								key[key_pos++] = p_data_buf[pos++];
						key[key_pos++] = 0;

						while(IS_BLANK(p_data_buf[pos]))
							pos++;

						while( !IS_BLANK(p_data_buf[pos])
							&& !IS_RETURN(p_data_buf[pos])
							&& val_pos < 255 )
							value[val_pos++] = p_data_buf[pos++];
						value[val_pos++] = 0;

						param_list_add_multi(&p_list, key, value, val_pos+1, 1);
					}
					// Interpret <</***(...)>>setpagedevice.
					else if( strncmp(p_data_buf, PAGE_DEV_BEGIN, dev_len) == 0
					 && (data_byte - 1 - end_len) )
					{
						if( strncmp(&p_data_buf[data_byte - 1 - end_len],
							 PAGE_DEV_END, end_len) == 0 )
						{
							int pos = dev_len;

							while( !IS_RETURN(p_data_buf[pos]) )
							{
								int key_pos = 0;
								int val_pos = 0;

								while( p_data_buf[pos] != '/'
									&& !IS_RETURN(p_data_buf[pos]) ) pos++;

								if( p_data_buf[pos] == '/' ) pos++;
								else continue;

								while( isalnum(p_data_buf[pos])
									&& key_pos < 255 )
										key[key_pos++] = p_data_buf[pos++];
								key[key_pos++] = 0;

								if( p_data_buf[pos] == '(' )
								{
									pos++;

									while( p_data_buf[pos] != ')'
										&& !IS_BLANK(p_data_buf[pos])
										&& !IS_RETURN(p_data_buf[pos])
										&& val_pos < 255 )
										value[val_pos++] = p_data_buf[pos++];
									value[val_pos++] = 0;

									if( p_data_buf[pos] == ')' ) pos++;
									else continue;

									if( !strcmp(key, "CNPageSizeName") )
										strcpy(key, "PageSize");
								}
								else continue;

								param_list_add_multi(
									&p_list, key, value, val_pos+1, 1);
							}
						}
					}

					if( data_byte > 0 )
					{
						if( fwrite(p_data_buf, 1, data_byte, ps_fp)
														 != data_byte )
							goto write_error;
					}
					data_byte = 0;
				}
			}
			if( data_byte > 0 )
			{
				if( fwrite(p_data_buf, 1, data_byte, ps_fp) != data_byte )
					goto write_error;
			}
		}
	}

	if( p_read_buf != NULL )
		free(p_read_buf);
	if( p_data_buf != NULL )
		free(p_data_buf);

	if( *p_ifd  != 0 )
		close(*p_ifd);

	fseek(ps_fp, 0L, SEEK_SET);
	*p_ifd = fileno(ps_fp);

	return p_list;

write_error:
	if( p_list != NULL )
		param_list_free(p_list);

	if( *p_ifd  != 0 )
		close(*p_ifd);
	*p_ifd = -1;

	return NULL;
}

static
char *strtolower(char *str)
{
	unsigned char *cc = str;
	while( *cc )
	{
		*cc = tolower(*cc);
		cc++;
	}
	return str;
}

static
int is_borderless(char* size_name)
{
	int len = strlen(size_name);

	if( !strcmp(&size_name[len - 3], ".bl") )
	{
		size_name[len - 3] = 0;
		return 1;
	}
	else
		return 0;
}

static
void mark_ps_param_options(ppd_file_t *p_ppd, ParamList *p_param)
{
	int num = param_list_num(p_param);

	if( num > 0 )
	{
		cups_option_t *options
			= (cups_option_t*)malloc(num * sizeof(cups_option_t));
		if( options )
		{
			int i;
			for( i = 0 ; i < num ; i++ )
			{
				options[i].name = p_param->key;
				options[i].value = strtolower(p_param->value);
				p_param = p_param->next;
			}
			for( i = num - 1 ; i >= 0 ; i-- )
			{
				cupsMarkOptions(p_ppd, 1, &options[i]);
			}

			free(options);
		}
	}
}

static
void get_min_max_width_length(
	PpdToOptKey *p_table, cups_option_t *p_opt, int num_opt,
	long *p_minw_mm, long *p_maxw_mm, long *p_minl_mm, long *p_maxl_mm)
{
	*p_minw_mm = -1;
	*p_maxw_mm = -1;
	*p_minl_mm = -1;
	*p_maxl_mm = -1;

	while( num_opt-- > 0 )
	{
		if( !strcmp(p_opt->name, "CNMinWidth") )
			*p_minw_mm = atoi(p_opt->value);
		else if( !strcmp(p_opt->name, "CNMaxWidth") )
			*p_maxw_mm = atoi(p_opt->value);
		else if( !strcmp(p_opt->name, "CNMinLength") )
			*p_minl_mm = atoi(p_opt->value);
		else if( !strcmp(p_opt->name, "CNMaxLength") )
			*p_maxl_mm = atoi(p_opt->value);

		p_opt++;
	}
}

static
char *parse_product_name(char *ppd_product, char *product, int len)
{
	char *p_product = product;
	char *p_cc;
	int count = 0;

	memset(product, 0, len);
	p_cc = ppd_product;

	while( *p_cc == '(' )
		p_cc++;

	while( *p_cc != ')' && *p_cc != '\0' && count < len - 1 )
	{
		*p_product++ = *p_cc++;
		count++;
	}
	*p_product = '\0';

	return product;
}

static
char *make_params(cups_option_t *p_opt, int num_opt,
	ParamList *p_param, char *p_product, int len)
{
	ppd_file_t *p_ppd;
	char *cmd_buf = NULL;
	char *ppd_name = getenv("PPD");

	if( (p_ppd = ppdOpenFile(ppd_name)) != NULL )
	{
		ParamList *p_list = NULL;
		PpdToOptKey *p_opt_key_table = alloc_opt_key_table(ppd_name);
		PpdToOptKey *p_table = p_opt_key_table;
		ppd_choice_t *p_choice;
		ppd_size_t *p_size;
		int reso;
		long minw_mm, maxw_mm;
		long minl_mm, maxl_mm;
		char gs_cmd_buf[1024];
		char *bj_cmd_buf;

		if( p_opt_key_table != NULL )
		{
			ppdMarkDefaults(p_ppd);
			mark_ps_param_options(p_ppd, p_param);
			cupsMarkOptions(p_ppd, num_opt, p_opt);

			// Obtain resolution setting.
			p_choice = ppdFindMarkedChoice(p_ppd, "Resolution");
			reso = atoi(p_choice->choice);

			// Obtain page size setting.
			p_choice = ppdFindMarkedChoice(p_ppd, "PageSize");
			p_size = ppdPageSize(p_ppd, p_choice->choice);

			// Obtain product name.
			parse_product_name(p_ppd->product, p_product, len);

			// Obtain min max width and height.
			get_min_max_width_length(p_opt_key_table, p_opt, num_opt,
				&minw_mm, &maxw_mm, &minl_mm, &maxl_mm);

			sprintf(gs_cmd_buf,
	"gs -r%d -g%dx%d -q -dNOPROMPT -dSAFER -sDEVICE=ppmraw -sOutputFile=- -| ",
	reso, (int)(p_size->width * (float)reso / 72.0),
	  (int)(p_size->length * (float)reso / 72.0));

			while( p_table->ppd_key != NULL )
			{
				p_choice = ppdFindMarkedChoice(p_ppd, p_table->ppd_key);

				if( p_choice != NULL )
				{
					char choice[256];

					strcpy(choice, p_choice->choice);

					if( !strcmp(p_table->ppd_key, "PageSize")
					 && !strcmp(choice, "Custom") )
					{
						char pw[256];
						char pl[256];
						long pw_mm, pl_mm;

						pw_mm = (long)((double)p_size->width * 2540.0 / 72.0);
						pl_mm = (long)((double)p_size->length * 2540.0 / 72.0);

						if( minw_mm != -1 && pw_mm < minw_mm )
							pw_mm = minw_mm;
						if( maxw_mm != -1 && pw_mm > maxw_mm )
							pw_mm = maxw_mm;
						if( minl_mm != -1 && pl_mm < minl_mm )
							pl_mm = minl_mm;
						if( maxl_mm != -1 && pl_mm > maxl_mm )
							pl_mm = maxw_mm;

						sprintf(pw, "%ld", pw_mm);
						sprintf(pl, "%ld", pl_mm);

						param_list_add(&p_list, p_table->opt_key, "user", 5);
						param_list_add(&p_list,
							"--paperwidth", pw, strlen(pw) + 1);
						param_list_add(&p_list,
							"--paperheight", pl, strlen(pl) + 1);
					}
					else
					{
						if( !strcmp(p_table->ppd_key, "PageSize")
						 && is_borderless(choice) )
						{
							param_list_add(&p_list, "--borderless", "", 1);
						}

						param_list_add(&p_list, p_table->opt_key,
							choice, strlen(choice) + 1);
					}
				}
				p_table++;
			}

			while( num_opt-- > 0 )
			{
				char *opt_key = ppd_to_opt_key(p_opt_key_table, p_opt->name);

				if( opt_key != NULL )
				{
					char *value = p_opt->value;
					param_list_add(&p_list, opt_key, value, strlen(value) + 1);
				}
				p_opt++;
			}

			{
				char bbox_buf[256];
				float dd = 1.0 - 36.0 / (float)reso;
				int left   = (int)p_size->left;
				int top    = (int)(p_size->top + dd);
				int right  = left + (int)(p_size->right - p_size->left + dd);
				int bottom = top - (int)(p_size->top - p_size->bottom + dd);

				sprintf(bbox_buf, "%-d,%-d,%-d,%-d", left, bottom, right, top);

				param_list_add(&p_list, "--bbox",
						bbox_buf, strlen(bbox_buf) + 1);
			}

			bj_cmd_buf = make_filter_param_line(p_product, reso, p_list);

			if( bj_cmd_buf )
			{
				int buf_len = strlen(gs_cmd_buf) + strlen(bj_cmd_buf) + 1;

				if( (cmd_buf = (char*)malloc(buf_len)) != NULL )
				{
					strcpy(cmd_buf, gs_cmd_buf);
					strcat(cmd_buf, bj_cmd_buf);
				}
				free(bj_cmd_buf);
			}

			if( p_list != NULL )
				param_list_free(p_list);

			free_opt_key_table(p_opt_key_table);
		}
		ppdClose(p_ppd);
	}
	return cmd_buf;
}

int	main(int argc, char *argv[])
{
	cups_option_t *p_opt = NULL;
	ParamList *p_param = NULL;
	char *cmd_param;
	char *p_data_buf;
	int num_opt = 0;
	int filter_pid = -1;
	int ifd;
	int fds[2];
	char *ps_path;
	char product[256];
	char filter_path[256];
	char *filter_param[4];

#ifdef	DEBUG_SLEEP
	sleep(1000);
#endif

	setbuf(stderr, NULL);

	signal(SIGCHLD, SIG_DFL);
	signal(SIGPIPE, SIG_DFL);

	if( argc < 6 || argc > 7 )
	{
		fputs("ERROR: pstocanonbj illegal parameter number -\n", stderr);
		return 1;
	}

	if( argv[5] != NULL )
	{
		num_opt = cupsParseOptions(argv[5], 0, &p_opt);
		if( num_opt < 0 )
		{
			perror("ERROR: illegal option - ");
			return 1;
		}
	}

	if( argc == 7 )
	{
		if( (ifd = open(argv[6], O_RDONLY)) == -1 )
		{
			perror("ERROR: can't open file - ");
			return 1;
		}
	}
	else
	{
		ifd = 0;
	}

	p_param = get_ps_params(&ifd, &ps_path);
	if( ifd == -1 )
	{
		perror("ERROR: can't open data file - ");
		return 1;
	}

	cmd_param = make_params(p_opt, num_opt, p_param, product, 256);

	if( cmd_param == NULL )
	{
		if( ifd != 0 )
			close(ifd);

		perror("ERROR: can't make parameter - ");
		return 1;
	}

	strcpy(filter_path, SHELL_PATH);
	strcat(filter_path, SHELL_NAME);

	filter_param[0] = SHELL_NAME;
	filter_param[1] = "-c";
	filter_param[2] = cmd_param;
	filter_param[3] = NULL;

	p_data_buf = (char*)malloc(BUF_SIZE);

	if( p_data_buf != NULL )
	{
#ifdef	DEBUG_PS
		FILE *log_fp = fopen("/tmp/bj_data_log.ps", "w");
#endif
		filter_pid = exec_filter(filter_path, filter_param, 1, fds);

		for(;;)
		{
			int read_bytes = read(ifd, p_data_buf, BUF_SIZE);

			if( read_bytes > 0 )
			{
				write(fds[1], p_data_buf, read_bytes);
#ifdef	DEBUG_PS
				fwrite(p_data_buf, read_bytes, 1, log_fp);
#endif
			}
			else if( read_bytes <= 0 )
				break;
		}

		free(p_data_buf);

#ifdef	DEBUG_PS
		if( log_fp )	fclose(log_fp);
#endif
	}

	if( p_param != NULL )
		param_list_free(p_param);

	free(cmd_param);

	if( ifd != 0 )
	{
		close(ifd);
		unlink(ps_path);
	}

	close(fds[1]);

	if( filter_pid != -1 )
		waitpid(filter_pid, NULL, 0);

	return 0;
}

